/*
 * Copyright 2005-2008 NVIDIA Corporation.  All rights reserved.
 */

// ===================================================================
// ===================================================================
//
// Module: opt_sync.cxx
// Date: 2008/06/12
// Author: John A. Stratton
//
// ===================================================================
//
//
//
// ===================================================================
//
// Description: Revome redundant synchronization points from an SPMD
// function.  Initially targeted at the NVIDIA CUDA language, this 
// process models barrier synchronization points and shared memory 
// uses in an analysis similar to dead store elimination.  Hence, we 
// call the class "dead sync elimination" 
//
// ===================================================================
// ===================================================================

#include "opt_sync.h"
#include "errors.h"
#include "erbe.h"
#include "erglob.h"

// ==================================================================
// Get the current summary state of sync liveness at a whirl node

void 
DSyncE::SetState(WN* stmt, LIVE_STATE val)
{
  if(stmt == NULL)
    return;

  LIVE_STATE* state = (LIVE_STATE*)WN_MAP_Get(_live_wns, stmt);
	
  if(state == NULL)
  {
    state = CXX_NEW(LIVE_STATE[1], &_loc_pool);
    WN_MAP_Set(_live_wns, stmt, state);
  }

  *state = val;
}

// ==================================================================
// Get the current summary state of sync liveness at a whirl node

LIVE_STATE 
DSyncE::GetState(WN* stmt)
{
  if(stmt == NULL)
    return Sync_Unkwn;

  LIVE_STATE* state = (LIVE_STATE*)WN_MAP_Get(_live_wns, stmt);

  if(state != NULL)
    return *state;
	
  SetState(stmt, Sync_Unkwn);
  return Sync_Unkwn;
}


// ==================================================================
// Set the current summary state of sync liveness at the exit of a bb

void
DSyncE::SetBBState(BB_NODE* bb, LIVE_STATE val)
{
  LIVE_STATE * old_state =
    (LIVE_STATE*) _live_bbs->Get_val((POINTER)(INTPTR)bb);

  if(old_state == NULL)
  {
    old_state = CXX_NEW(LIVE_STATE, &_loc_pool);
    _live_bbs->Add_map(((POINTER)(INTPTR)bb), old_state);
  }

  *old_state = val;
}


// ==================================================================
// Get the current summary state of sync liveness at the exit of a bb

LIVE_STATE
DSyncE::GetBBState(BB_NODE* bb)
{
  LIVE_STATE * state = (LIVE_STATE*)_live_bbs->Get_val((POINTER)(INTPTR)bb);

  if(state == NULL)
    return Sync_Unkwn;

  return *state;
}

// ===================================================================
//
// Description:
//
// TRUE if the whirl node is a __syncthread() intrinsic, 
// FALSE otherwise
//
// ===================================================================
static BOOL 
Is_syncthreads(WN * wn)
{
  FmtAssert(wn != NULL, ("null node tested for __synchthreads()"));

  if (WN_operator(wn) == OPR_INTRINSIC_CALL) {
    return (WN_intrinsic(wn) == INTRN_SYNCHRONIZE);
  }
  return FALSE;
}

BOOL 
DSyncE::Has_Shared_Mem_Access(WN * root) 
{
  for (WN_ITER* wni = WN_WALK_TreeIter(root);
       wni != NULL;
       wni = WN_WALK_TreeNext(wni)) 
  {
    WN* addr = WN_ITER_wn(wni);
    switch(WN_operator(addr))
    {
      case OPR_ISTORE:
      case OPR_ILOAD:
	//Assume worst case
	return TRUE;
      case OPR_LDID:
      case OPR_STID:
      case OPR_LDBITS:
      case OPR_STBITS:
	//Check for shared memory access
	ST* real_sym;
	//Check the auxillary symbol table if necessary
	if(WN_has_aux(addr))
	  real_sym = Opt_stab()->St_ptr(addr);
	else
	  real_sym = WN_st(addr);

	if(real_sym == NULL)
	  return TRUE;
	if(ST_in_shared_mem(real_sym))
	  return TRUE;

	break;
      //Cover the COMIC case, where smem is represented by intrinsics
      case OPR_INTRINSIC_OP:
	if( WN_intrinsic(addr) == INTRN_GETSHAREDMEM)
	  return TRUE;
      default:
	break;
    }
  }

  return FALSE;
}

// Removes and deletes a wn from a doubly linked list of wns in a bb
void
Remove_WN_from_BB(WN* wn, BB_NODE* bb)
{
  if(wn == NULL)
    return;
  if(bb == NULL) {
    WN_Delete(wn);
    return;
  }

  if(WN_next(wn) != NULL)
    WN_prev(WN_next(wn)) = WN_prev(wn);
  else
    bb->Set_laststmt(WN_prev(wn));

  if(WN_prev(wn) != NULL)
    WN_next(WN_prev(wn)) = WN_next(wn);
  else
    bb->Set_firststmt(WN_next(wn));

  WN_Delete(wn);
}


DSyncE::DSyncE(CFG* cfg, OPT_STAB* opt_stab):
  _cfg(cfg), _opt_stab(opt_stab)
{
  MEM_POOL_Initialize(&_loc_pool, "dead_sync_pool", FALSE);
  MEM_POOL_Push(&_loc_pool);
  _live_wns = WN_MAP_Create(&_loc_pool);
  _live_bbs = CXX_NEW(MAP(101, &_loc_pool), &_loc_pool);
}

DSyncE::~DSyncE()
{
  WN_MAP_Delete(_live_wns);
  CXX_DELETE(_live_bbs, &_loc_pool);
  MEM_POOL_Pop(&_loc_pool);
  MEM_POOL_Delete(&_loc_pool);
}


// ===================================================================
// This propogates the liveness information of sync nodes through the 
// reverse cfg.  It is assumed that each sync stmt and the terminal 
// statement of each bb is already marked with the summary liveness 
// state at that point. 
// ===================================================================

LIVE_STATE
DSyncE::Propogate_Liveness(WN* sync, BB_NODE* bb)
{
  FmtAssert(bb != NULL, ("traversal bb is null"));
  LIVE_STATE initial_state;

  //For the first level call, check the sync node's state 
  //rather than the basic block's state
  if(sync != NULL)
    initial_state = GetState(sync);
  else
    initial_state = GetBBState(bb);

  // If we already know the current state, return it
  if(initial_state != Sync_Unkwn)
    return initial_state;

  // Before traversing, mark the current node as dead, so we don't 
  // get infinite recursion on cycles, and get the correct information.
  // For the first level of recursion, the bb -must- have a non-unknown 
  // state, since it contains a sync point.
  if(sync == NULL)
    SetBBState(bb, Sync_Dead);

  BB_NODE* pred;
  BB_LIST_ITER pred_iter;
  LIVE_STATE path_state = Sync_Dead;

  FOR_ALL_ELEM( pred, pred_iter, Init(bb->Pred()))
  {
    // All deeper levels get a NULL sync pointer so they examine 
    // the end-block state
    path_state = Propogate_Liveness(NULL, pred);

    if(path_state == Sync_Live)
      break;
  }

  // If the for loop exited without finding a live path, the 
  // path_state will fall through with a dead value.
  // If this is the first level, the liveness leaving this bb 
  // is already defined, and independent of the liveness entering
  if(sync == NULL)
    SetBBState(bb, path_state);

  return path_state;
}


// ===================================================================
// Perform dead sync elimination
// ===================================================================

void
DSyncE::Do_dead_sync_elim()
{
  STACK<WN*> worklist_wn(&_loc_pool);
  STACK<BB_NODE*> worklist_bb(&_loc_pool);

  //Compute summary liveness information for each BB while searching 
  //for sync statements. 
  CFG_ITER cfgi(Cfg());
  BB_NODE* node;


  FOR_ALL_NODE(node, cfgi, Init())
  {
    WN* wn;
    WN* current_sync = NULL;

    LIVE_STATE current_state = Sync_Unkwn;

    for(wn = node->Laststmt(); wn != NULL; wn = WN_prev(wn)) 
    {
      // Synchronization kills, and with no intervening statements 
      // to redefine liveness, the last stmt in this region is sync-dead.
      if(Is_syncthreads(wn))
      {
	if(current_state == Sync_Unkwn)
	  current_state = Sync_Dead;
	
	if(current_sync != NULL)
	  SetState(current_sync, current_state);
	else
	  SetBBState(node, current_state);
	current_sync = wn;

	// Save the sync point to propogate in the future
	worklist_wn.Push(wn);
	worklist_bb.Push(node);
	current_state = Sync_Unkwn;
      }
      // We define all shared memory accesses as defining sync liveness.
      else if(Has_Shared_Mem_Access(wn)) 
	current_state = Sync_Live;
    }

    // Save whatever the final state for the last explored stmt of 
    // this bb was.
    if(current_sync == NULL)
      SetBBState(node, current_state);
    else
      SetState(current_sync, current_state);
  }

  //For each synchronization point, propogate its deadness/liveness.
  while(worklist_wn.Elements() != 0)
  {
    WN* sync = worklist_wn.Pop();
    BB_NODE* start_block = worklist_bb.Pop();

    LIVE_STATE final_state = Propogate_Liveness(sync, start_block);

    // If the final state is dead, take out the sync.  
    if(final_state == Sync_Dead)
    {
      if ((sync != NULL) && OPERATOR_is_stmt(WN_operator(sync)))
        ErrMsgSrcpos(EC_Dead_Sync, WN_Get_Linenum(sync));
      else
        ErrMsg(EC_Dead_Sync);

      //start_block->Remove_stmtrep(sync_rep);
      Remove_WN_from_BB(sync, start_block);
    }
  }
}
  







  




